from ._impl.procedure import Procedure

# Error Strategies
from ._impl.strategies.base import BaseStrategy
from ._impl.strategies.blacklist import BlacklistStrategy
from ._impl.strategies.whitelist import WhitelistStrategy
from ._impl.strategies.donthandle import DontHandleStrategy
from ._impl.strategies.justhandle import JustHandleStrategy
from ._impl.strategies.returnerror import ReturnErrorStrategy


# Retry Policies
from ._impl.policies.base import BasePolicy
from ._impl.policies.eventbased import EventBasedPolicy
from ._impl.policies.idempotent import IdempotentPolicy
from ._impl.policies.timebased import TimeBasedPolicy


# UTILS
from .abtract import AbstractCore

# TYPING
from abc import abstractmethod as AbstractMethod
from typing import Any


class FunnelCore(AbstractCore):
    @AbstractMethod
    def aggregate(*records: Any) -> Any: ...
    


# UTILS
from .abtract import AbstractCore
from ._functools import MultiFunction, CoreNamespace

# TYPING
from abc import (
    ABCMeta as AbstractMetadata,
)

# TODO: Make sure after all the processor functions are cleared and renamed into 'process', user should see only that one instead of original names

class ProcessMetaCore(AbstractMetadata):
    def __init__(cls, name, bases, clsdict):
        super().__init__(name, bases, clsdict)
    
    def __new__(cls, name, bases, dct):
        
        # Skip below manipulations on parent class
        if AbstractCore in bases:
            return super().__new__(cls, name, bases, dct)
        
        # Merge all processor methods into single object
        accumulative = None
        for key in dct.copy():
            attr = dct[key]
            # Get those methods that are marked as processor
            if accumulative == None and isinstance(attr, MultiFunction):
                accumulative = attr
                
            # Clean up the rest of the class namespace
            if key not in ['__module__', '__qualname__', '__class__']:
                del dct[key]
                
        # Add newly formed accumulative processor as attribute
        dct['process'] = accumulative
        return super().__new__(cls, name, bases, dct)

def processor(func):
    if not hasattr(processor, 'namespaces'):
        processor.namespaces = {}
    
    module_name = func.__qualname__.split('.')[0]
    if not module_name in processor.namespaces:
        processor.namespaces[module_name] = CoreNamespace()
    return processor.namespaces[module_name].register(func)

class ProcessCore(AbstractCore, metaclass=ProcessMetaCore):
    ...
    
    


from typing import Callable

def clean_raise(exception: BaseException):
    exception.__traceback__ = None
    raise exception from None

def fn2key(fn):
    return tuple([
            fn.__module__,
            fn.__class__,
            fn.__name__,
        ])

class MultiFunction(object):
    def __init__(self, fn = None):
        self.functions = list()
        
        if isinstance(fn, MultiFunction):
            self.extend(fn)
        elif isinstance(fn, Callable):
            self.functions.append(fn)
        
    def __call__(self, *args, **kwargs):
        result = None
        
        for fn in self.functions:
            result = fn(*args, **kwargs)
            args = (result,)

        return result
    
    
    def __add__(self, new_fn):
        if not isinstance(new_fn, Callable):
            raise TypeError("Can't append anything other than object of type Callable.")
        
        if isinstance(new_fn, MultiFunction):
            self.functions.extend(new_fn.functions)
        else:
            self.functions.append(new_fn)
        
        return self
 
    def __str__(self) -> str:
        return str(self.functions)
        
    def key(self):
        fn = self.functions[0] if len(self.functions) > 0 else None
        
        if fn:
            return fn2key(fn)
        else:
            return tuple()


class CoreNamespace:
    def __init__(self) -> None:
        self.functions = MultiFunction()

    def register(self, fn):
        self.functions += fn
        return self.functions



from ...cores import IngressCore
from .abtract import AbstractNode

# TYPING
from typing import (
    Iterable,
    Any,
    NoReturn,
    Union,
)

class IngressNode(AbstractNode):
    def __init__(self, ingress_core: IngressCore) -> NoReturn:
        super().__init__(ingress_core)
    
    @property
    def input(self) -> Iterable:
        return self.core.input
    
    @input.setter
    def input(self, new_input: Union[Iterable, Any]):
        self.core.input = new_input

    @property
    def output(self):
        return self.core


# UTILS
import itertools

# TYPING
from abc import (
    ABC as AbstractClass,
    ABCMeta as AbstractMetadata,
    abstractmethod as AbstractMethod
)
from typing import (
    Iterable,
    Union,
    Any,
)



class AbstractCore(AbstractClass):
    @property
    def input(self) -> Iterable:
        return self._input
    
    @input.setter
    def input(self, new_input: Union[Iterable, Any]):
        if not isinstance(new_input, Iterable):
            new_input = [new_input]
            
        self._input = itertools.chain(new_input)


# UTILS
from ._functools import clean_raise
from .abtract import AbstractCore
import itertools
import time

# TYPING
from abc import abstractmethod as AbstractMethod

from typing import (
    Iterable,
    Any,
    NoReturn,
    Union
)


class EgressCore(AbstractCore):
    
    def __init__(self, input:Union[Iterable, Any] = None, heartbeat:int = 1, forgiving=False, backpressure:int = 0) -> NoReturn:
        self.input = input
        self.heartbeat = heartbeat
        self.forgiving = forgiving
        self.backpressure = backpressure

    def __enter__(self):
        self.constructor()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.destructor(exc_type, exc_value, traceback)
        
    def run(self) -> NoReturn:
        i = 0
        for record in self.input:
            i += 1
            any_exception = None
            try:
                self.consume(record)
            except Exception as e:
                any_exception = e
            finally:
                # If callback recieves None as parameter
                self.callback(record, any_exception)

                if self.forgiving == False and any_exception is not None:
                    clean_raise(any_exception)
                
                if i % self.heartbeat == 0:
                    self.pulse()
                    time.sleep(self.backpressure)
        else:
            # Run pulse function one last time 
            # for the remaining records where i % self.heartbeat > 0
            if i % self.heartbeat > 0:
                self.pulse()
                
    @property
    def input(self) -> Iterable:
        return self._input
    
    @input.setter
    def input(self, new_input):
        if not isinstance(new_input, Iterable):
            new_input = itertools.chain([new_input])
        
        self._input = new_input
        
    @AbstractMethod
    def constructor(self) -> NoReturn: ...
    @AbstractMethod
    def destructor(self, exc_type, exc_value, traceback) -> NoReturn: ...
    @AbstractMethod
    def callback(self, record: Any, exception: Exception) -> NoReturn: ...
    @AbstractMethod
    def pulse(self) -> NoReturn: ...
    @AbstractMethod
    def consume(self, record: Any) -> NoReturn: ... 



from ._impl.abtract import AbstractCore
from ._impl.ingress import IngressCore, IngressType
from ._impl.process import ProcessCore, processor
from ._impl.egress import EgressCore
from ._impl.junction import JunctionCore
from ._impl.funnel import FunnelCore
    
    
    


from ...cores import JunctionCore
from .abtract import AbstractNode, Shiftable
from collections import deque

import itertools

# TYPING
from typing import (
    Iterable,
    NoReturn,
)

class Outflow:
    def __init__(self, source: Iterable, outputs:int=2) -> None:
        self.source = source
        self.outputs = [deque() for _ in range(outputs)]

    def iterate(self):
        values = next(self.source)
        for i, value in enumerate(values):
            self.outputs[i].append(value)

    def next(self, i):
        if len(self.outputs[i]) == 0:
            self.iterate()

        return self.outputs[i].popleft()


class Flow(Shiftable):
    def __init__(self, outflow: Outflow, index: int) -> None:
        self.outflow = outflow
        self.index = index

    def __iter__(self):
        return self

    def __next__(self):
        return self.outflow.next(self.index)
    
    @property
    def output(self) -> Iterable:
        return self

class JunctionNode(AbstractNode):
    def __init__(self, junction_core: JunctionCore, outflows = 2, input = None) -> NoReturn:
        self.outflows = outflows
        self.input = input
        super().__init__(junction_core)
        
    def __getitem__(self, key):
        return Flow(self.__input, key)
    
    def __iter__(self):
        return (Flow(self.__input, i) for i in range(self.outflows))
    
    @property
    def input(self):
        return super().input
    
    @input.setter
    def input(self, new_input):
        if not isinstance(new_input, Iterable):
            new_input = [new_input]
            
        self.__input = itertools.chain(new_input)
        self.__input = map(lambda record: self.core.segregate(record), self.__input)
        self.__input = Outflow(self.__input, self.outflows)
    
    def __rshift__(self, other: 'AbstractNode'):
        raise TypeError("JunctionNode can only pass output from it's slices.")        
    
   


def void_callback(*args, **kwargs): ...


def void_execute(self):
    raise StopIteration("Ingress Core has finished.")


def void_available(self):
    return True


def void_iterate(self):
    ...


def void_input_setter(self, input):
    return input


def no_change(x): 
    return x



from ...cores import ProcessCore
from .abtract import AbstractNode

# TYPING
from typing import NoReturn, Iterable

class ProcessNode(AbstractNode):
    def __init__(self, *processor_cores: ProcessCore, input: Iterable = None) -> NoReturn:
        super().__init__(processor_cores, input = input)
    
    def __iter__(self):
        return self.output
        
    @property
    def output(self):
        # SIMPLER: return map(self.core.process, self.input)
        transformed_output = self.input
        for acore in self.core:
            transformed_output = map(acore.process, transformed_output)
        
        return transformed_output



# UTILS
from ._templates import void_available, void_iterate
from .abtract import AbstractCore
from enum import Enum
import itertools

# TYPING
from abc import (
    ABCMeta as AbstractMetadata,
    abstractmethod as AbstractMethod
)
from typing import (
    Iterable,
    Any,
    NoReturn,
    Generator,
    Union
)


class IngressType(Enum):
    # Input based Ingress: user sets a seed in form of a iterable
    # Process finishes when the given input gets exhausted
    # Kind of deterministic approach where you know when the process is going to be over.
    INPUT = 0x10

    # Function based Ingress: user defines certain function that will be executed in certain order.
    # Process finishes when flag function returns False
    # Kind of probabilitic or indefinite approach, where the process completely depends on implementation of function (maybe external api's behavior or kafka's consuming)
    FUNCTION = 0x20


class IngressMetaCore(AbstractMetadata):
    def __new__(cls, name, bases, dct, Type: IngressType = None):
        if Type is not None and not isinstance(Type, IngressType):
            raise TypeError("Type parameter can only be of type IngressType.")
        dct['_type'] = Type
        return super().__new__(cls, name, bases, dct)

class IterationHandler:
    class Defaults(Enum):
        NULL_VALUE = 0x123
        STOP_VALUE = 0x234
    
    def __init__(self) -> None:
        self.iterators = itertools.chain([])
        
    def append(self, genexpr, flatten = False):
        if isinstance(genexpr, Generator):
            self.iterators = itertools.chain(genexpr, self.iterators) # TODO: see if there is any error here
        else:
            if flatten:
                self.iterators = itertools.chain(self.iterators, genexpr)
            else:
                self.iterators = itertools.chain(self.iterators, [genexpr])
    
    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterators)
                    
class IngressCore(AbstractCore, metaclass=IngressMetaCore):
    Types = IngressType
    _type: IngressType = None
    
    def __init__(self, input: Union[Iterable, Any] = None, flatten = False) -> NoReturn:
        self.flatten = flatten
        self.input = input
        self.__first_run = True
    
    def __super_constructor(self):
        self.handler = IterationHandler()
    
    def __super_destructor(self, exc_type, exc_value, traceback):
        ...
        
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if cls._type == IngressType.INPUT:
            setattr(cls, 'available', void_available)
            setattr(cls, 'iterate', void_iterate)
        elif cls._type == IngressType.FUNCTION:
            ...
        else:
            raise AttributeError("When extending 'IngressCore' specify 'Type' as meta variable of type 'IngressType'")
    
    def __iter__(self):
        return self

    def __next__(self):
        if self.__first_run:
            if self._type == IngressType.FUNCTION:
                self.handler.append(self.produce(), self.flatten) # Add first batch
            self.__first_run = False

        while True:
            try:
                # Exhaust Input Handler first
                return next(self.handler)
            except StopIteration:
                if self.available():
                    self.iterate()
                else:
                    # If no more available stop all
                    raise StopIteration 
                
                # If handler exhausted then try to load
                self.handler.append(self.produce(), self.flatten)
                
            except RuntimeError as e:
                if str(e) == 'generator raised StopIteration':
                    raise StopIteration
            

    def __enter__(self):
        self.__super_constructor()
        self.constructor()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__super_destructor(exc_type, exc_value, traceback)
        self.destructor(exc_type, exc_value, traceback)
    
    @AbstractMethod
    def available(self) -> bool: ...

    @AbstractMethod
    def iterate(self) -> NoReturn: ...
    
    # @AbstractMethod
    def constructor(self): ...

    # @AbstractMethod
    def destructor(self, exc_type, exc_value, traceback): ...
    
    @AbstractMethod
    def produce(self) -> Union[Iterable, Any]: ... 


# UTILS
from .abtract import AbstractCore

# TYPING
from abc import abstractmethod as AbstractMethod
from typing import Any, Tuple

class JunctionCore(AbstractCore):
    @AbstractMethod
    def segregate(record: Any) -> Tuple[Any]: ...



# TOOLS
from .abtract import AbstractNode
from ...cores import EgressCore

# TYPING
from typing import (
    Iterable,
    Any,
    NoReturn,
    Iterable,
    Union
)

class EgressNode(AbstractNode):
    def __init__(self, egress_core: EgressCore) -> NoReturn:
        super().__init__(egress_core)

    def run(self):
        self.core.run()
        
    @property
    def input(self) -> Iterable:
        return self.core.input
    
    @input.setter
    def input(self, new_input: Union[Iterable, Any]):
        self.core.input = new_input


# TOOLS
from ...cores import AbstractCore

# TYPING
from abc import (
    ABC as AbstractClass,
    ABCMeta as AbstractMetadata,
)
from typing import (
    Iterable,
    Any,
    NoReturn,
    Union
)
# Iterable | Any

class Shiftable:
    def __rshift__(self: 'AbstractNode', other: 'AbstractNode'):
        other.input = self.output
        return other
    
    def __lshift__(self: 'AbstractNode', other: 'AbstractNode'):
        self.input = other.output
        return self

class AbstractNode(AbstractClass, Shiftable, metaclass=AbstractMetadata):
    def __init__(self, core: AbstractCore, input = None) -> NoReturn:
        self.core = core
        # init placeholder variables
        self.__input=input
        self.__output=None
        
    @property
    def input(self) -> Iterable:
        return self.__input
    
    @input.setter
    def input(self, value: Union[Iterable, Any]):
        self.__input = value

    @property
    def output(self) -> Iterable:
        return self.__output

    def __iter__(self):
        return self.output
        
    
    # @output.setter
    # def output(self, value):
    #     self.__output = value
        
   


from ._impl.cores.ingress import IngressApify
from ._impl.cores.egress import KafkaEgressCore, Sqlite3EgressCore
from ._impl.cores.junction import DuplicateRecordCore
from ._impl.cores.process import AnyJSONEncodeCore

from ._impl.nodes.junction import DuplicateRecordNode
from ._impl.nodes.process import AnyJSONEncodeNode


from ...cores import FunnelCore
from .abtract import AbstractNode
import itertools

# TYPING
from typing import (
    Iterable,
    Any,
    NoReturn,
    Generator,
    List,
    Union
)

def conjoined_enumerate(generators) -> Generator[List, None, None]:
    """_summary_
    Example Use Case:
    Aggregate:  +
    Inputs:     [1,2], [3,0, 99], []
    Output:     [1 + 3 + None, 2 + 0 + None, None + 99 + None]
    
    Args:
        generators (_type_): _description_

    Yields:
        Generator[List, None, None]: _description_
    """
    while True:
        next_elements = []
        all_done = True
        for gen in generators:
            try:
                next_element = next(gen)
                next_elements.append(next_element)
                all_done = False
            except StopIteration:
                next_elements.append(None)
        if all_done or all(elem is None for elem in next_elements):
            break
        
        yield next_elements

class FunnelNode(AbstractNode): 
    def __init__(self, funnel_core: FunnelCore) -> NoReturn:
        self.input_flows = list()
        super().__init__(funnel_core)

    @property
    def output(self):
        conjoined_flows = conjoined_enumerate(self.input_flows)
        return map(lambda flow: self.core.aggregate(*flow), conjoined_flows)

    ############################################################
    # Node Operations
    ############################################################

    @property
    def input(self) -> List[Iterable]:
        return self.input_flows
    
    @input.setter
    def input(self, new_input: Union[Iterable, Any]):
        if not isinstance(new_input, Iterable):
            new_input = [new_input]
            
        self.input_flows.append(itertools.chain(new_input))
        

    def __add__(self, other: AbstractNode):
        self.input = other.output
        return self
